from .values import *

class protocol_packet_handler(object):
    def __init__(self, portHandler):
        #self.sts_setend(protocol_end)# STServo bit end(STS/SMS=0, SCS=1)
        self.portHandler = portHandler
        self.sts_end = 0

    def sts_getend(self):
        return self.sts_end

    def sts_setend(self, e):
        self.sts_end = e

    def sts_tohost(self, a, b):
        if (a & (1<<b)):
            return -(a & ~(1<<b))
        else:
            return a

    def sts_toscs(self, a, b):
        if (a<0):
            return (-a | (1<<b))
        else:
            return a

    def sts_makeword(self, a, b):
        if self.sts_end==0:
            return (a & 0xFF) | ((b & 0xFF) << 8)
        else:
            return (b & 0xFF) | ((a & 0xFF) << 8)

    def sts_makedword(self, a, b):
        return (a & 0xFFFF) | (b & 0xFFFF) << 16

    def sts_loword(self, l):
        return l & 0xFFFF

    def sts_hiword(self, h):
        return (h >> 16) & 0xFFFF

    def sts_lobyte(self, w):
        if self.sts_end==0:
            return w & 0xFF
        else:
            return (w >> 8) & 0xFF

    def sts_hibyte(self, w):
        if self.sts_end==0:
            return (w >> 8) & 0xFF
        else:
            return w & 0xFF
        
    def getProtocolVersion(self):
        return 1.0

    def getTxRxResult(self, result):
        if result == COMM_SUCCESS:
            return "[TxRxResult] Communication success!"
        elif result == COMM_PORT_BUSY:
            return "[TxRxResult] Port is in use!"
        elif result == COMM_TX_FAIL:
            return "[TxRxResult] Failed transmit instruction packet!"
        elif result == COMM_RX_FAIL:
            return "[TxRxResult] Failed get status packet from device!"
        elif result == COMM_TX_ERROR:
            return "[TxRxResult] Incorrect instruction packet!"
        elif result == COMM_RX_WAITING:
            return "[TxRxResult] Now receiving status packet!"
        elif result == COMM_RX_TIMEOUT:
            return "[TxRxResult] There is no status packet!"
        elif result == COMM_RX_CORRUPT:
            return "[TxRxResult] Incorrect status packet!"
        elif result == COMM_NOT_AVAILABLE:
            return "[TxRxResult] Protocol does not support this function!"
        else:
            return ""

    def getRxPacketError(self, error):
        if error & ERRBIT_VOLTAGE:
            return "[ServoStatus] Input voltage error!"

        if error & ERRBIT_ANGLE:
            return "[ServoStatus] Angle sen error!"

        if error & ERRBIT_OVERHEAT:
            return "[ServoStatus] Overheat error!"

        if error & ERRBIT_OVERELE:
            return "[ServoStatus] OverEle error!"
        
        if error & ERRBIT_OVERLOAD:
            return "[ServoStatus] Overload error!"

        return ""

    def txPacket(self, txpacket):
        checksum = 0
        total_packet_length = txpacket[PKT_LENGTH] + 4  # 4: HEADER0 HEADER1 ID LENGTH

        if self.portHandler.is_using:
            return COMM_PORT_BUSY
        self.portHandler.is_using = True

        # check max packet length
        if total_packet_length > TXPACKET_MAX_LEN:
            self.portHandler.is_using = False
            return COMM_TX_ERROR

        # make packet header
        txpacket[PKT_HEADER_0] = 0xFF
        txpacket[PKT_HEADER_1] = 0xFF

        # add a checksum to the packet
        for idx in range(2, total_packet_length - 1):  # except header, checksum
            checksum += txpacket[idx]

        txpacket[total_packet_length - 1] = ~checksum & 0xFF

        #print "[TxPacket] %r" % txpacket

        # tx packet
        self.portHandler.clearPort()
        written_packet_length = self.portHandler.writePort(txpacket)
        if total_packet_length != written_packet_length:
            self.portHandler.is_using = False
            return COMM_TX_FAIL

        return COMM_SUCCESS

    def rxPacket(self):
        rxpacket = []

        result = COMM_TX_FAIL
        checksum = 0
        rx_length = 0
        wait_length = 6  # minimum length (HEADER0 HEADER1 ID LENGTH ERROR CHKSUM)

        while True:
            rxpacket.extend(self.portHandler.readPort(wait_length - rx_length))
            rx_length = len(rxpacket)
            if rx_length >= wait_length:
                # find packet header
                for idx in range(0, (rx_length - 1)):
                    if (rxpacket[idx] == 0xFF) and (rxpacket[idx + 1] == 0xFF):
                        break

                if idx == 0:  # found at the beginning of the packet
                    if (rxpacket[PKT_ID] > 0xFD) or (rxpacket[PKT_LENGTH] > RXPACKET_MAX_LEN) or (
                            rxpacket[PKT_ERROR] > 0x7F):
                        # unavailable ID or unavailable Length or unavailable Error
                        # remove the first byte in the packet
                        del rxpacket[0]
                        rx_length -= 1
                        continue

                    # re-calculate the exact length of the rx packet
                    if wait_length != (rxpacket[PKT_LENGTH] + PKT_LENGTH + 1):
                        wait_length = rxpacket[PKT_LENGTH] + PKT_LENGTH + 1
                        continue

                    if rx_length < wait_length:
                        # check timeout
                        if self.portHandler.isPacketTimeout():
                            if rx_length == 0:
                                result = COMM_RX_TIMEOUT
                            else:
                                result = COMM_RX_CORRUPT
                            break
                        else:
                            continue

                    # calculate checksum
                    for i in range(2, wait_length - 1):  # except header, checksum
                        checksum += rxpacket[i]
                    checksum = ~checksum & 0xFF

                    # verify checksum
                    if rxpacket[wait_length - 1] == checksum:
                        result = COMM_SUCCESS
                    else:
                        result = COMM_RX_CORRUPT
                    break

                else:
                    # remove unnecessary packets
                    del rxpacket[0: idx]
                    rx_length -= idx

            else:
                # check timeout
                if self.portHandler.isPacketTimeout():
                    if rx_length == 0:
                        result = COMM_RX_TIMEOUT
                    else:
                        result = COMM_RX_CORRUPT
                    break

        self.portHandler.is_using = False
        return rxpacket, result

    def txRxPacket(self, txpacket):
        rxpacket = None
        error = 0

        # tx packet
        result = self.txPacket(txpacket)
        if result != COMM_SUCCESS:
            return rxpacket, result, error

        # (ID == Broadcast ID) == no need to wait for status packet or not available
        if (txpacket[PKT_ID] == BROADCAST_ID):
            self.portHandler.is_using = False
            return rxpacket, result, error

        # set packet timeout
        if txpacket[PKT_INSTRUCTION] == INST_READ:
            self.portHandler.setPacketTimeout(txpacket[PKT_PARAMETER0 + 1] + 6)
        else:
            self.portHandler.setPacketTimeout(6)  # HEADER0 HEADER1 ID LENGTH ERROR CHECKSUM

        # rx packet
        while True:
            rxpacket, result = self.rxPacket()
            if result != COMM_SUCCESS or txpacket[PKT_ID] == rxpacket[PKT_ID]:
                break

        if result == COMM_SUCCESS and txpacket[PKT_ID] == rxpacket[PKT_ID]:
            error = rxpacket[PKT_ERROR]

        return rxpacket, result, error

    def ping(self, sts_id):
        model_number = 0
        error = 0

        txpacket = [0] * 6

        if sts_id >= BROADCAST_ID:
            return model_number, COMM_NOT_AVAILABLE, error

        txpacket[PKT_ID] = sts_id
        txpacket[PKT_LENGTH] = 2
        txpacket[PKT_INSTRUCTION] = INST_PING

        rxpacket, result, error = self.txRxPacket(txpacket)

        if result == COMM_SUCCESS:
            data_read, result, error = self.readTxRx(sts_id, 3, 2)  # Address 3 : Model Number
            if result == COMM_SUCCESS:
                model_number = self.sts_makeword(data_read[0], data_read[1])

        return model_number, result, error

    def action(self, sts_id):
        txpacket = [0] * 6

        txpacket[PKT_ID] = sts_id
        txpacket[PKT_LENGTH] = 2
        txpacket[PKT_INSTRUCTION] = INST_ACTION

        _, result, _ = self.txRxPacket(txpacket)

        return result

    def readTx(self, sts_id, address, length):

        txpacket = [0] * 8

        if sts_id >= BROADCAST_ID:
            return COMM_NOT_AVAILABLE

        txpacket[PKT_ID] = sts_id
        txpacket[PKT_LENGTH] = 4
        txpacket[PKT_INSTRUCTION] = INST_READ
        txpacket[PKT_PARAMETER0 + 0] = address
        txpacket[PKT_PARAMETER0 + 1] = length

        result = self.txPacket(txpacket)

        # set packet timeout
        if result == COMM_SUCCESS:
            self.portHandler.setPacketTimeout(length + 6)

        return result

    def readRx(self, sts_id, length):
        result = COMM_TX_FAIL
        error = 0

        rxpacket = None
        data = []

        while True:
            rxpacket, result = self.rxPacket()

            if result != COMM_SUCCESS or rxpacket[PKT_ID] == sts_id:
                break

        if result == COMM_SUCCESS and rxpacket[PKT_ID] == sts_id:
            error = rxpacket[PKT_ERROR]

            data.extend(rxpacket[PKT_PARAMETER0 : PKT_PARAMETER0+length])

        return data, result, error

    def readTxRx(self, sts_id, address, length):
        txpacket = [0] * 8
        data = []

        if sts_id >= BROADCAST_ID:
            return data, COMM_NOT_AVAILABLE, 0

        txpacket[PKT_ID] = sts_id
        txpacket[PKT_LENGTH] = 4
        txpacket[PKT_INSTRUCTION] = INST_READ
        txpacket[PKT_PARAMETER0 + 0] = address
        txpacket[PKT_PARAMETER0 + 1] = length

        rxpacket, result, error = self.txRxPacket(txpacket)
        if result == COMM_SUCCESS:
            error = rxpacket[PKT_ERROR]

            data.extend(rxpacket[PKT_PARAMETER0 : PKT_PARAMETER0+length])

        return data, result, error

    def read1ByteTx(self, sts_id, address):
        return self.readTx(sts_id, address, 1)

    def read1ByteRx(self, sts_id):
        data, result, error = self.readRx(sts_id, 1)
        data_read = data[0] if (result == COMM_SUCCESS) else 0
        return data_read, result, error

    def read1ByteTxRx(self, sts_id, address):
        data, result, error = self.readTxRx(sts_id, address, 1)
        data_read = data[0] if (result == COMM_SUCCESS) else 0
        return data_read, result, error

    def read2ByteTx(self, sts_id, address):
        return self.readTx(sts_id, address, 2)

    def read2ByteRx(self, sts_id):
        data, result, error = self.readRx(sts_id, 2)
        data_read = self.sts_makeword(data[0], data[1]) if (result == COMM_SUCCESS) else 0
        return data_read, result, error

    def read2ByteTxRx(self, sts_id, address):
        data, result, error = self.readTxRx(sts_id, address, 2)
        data_read = self.sts_makeword(data[0], data[1]) if (result == COMM_SUCCESS) else 0
        return data_read, result, error

    def read4ByteTx(self, sts_id, address):
        return self.readTx(sts_id, address, 4)

    def read4ByteRx(self, sts_id):
        data, result, error = self.readRx(sts_id, 4)
        data_read = self.sts_makedword(self.sts_makeword(data[0], data[1]),
                                  self.sts_makeword(data[2], data[3])) if (result == COMM_SUCCESS) else 0
        return data_read, result, error

    def read4ByteTxRx(self, sts_id, address):
        data, result, error = self.readTxRx(sts_id, address, 4)
        data_read = self.sts_makedword(self.sts_makeword(data[0], data[1]),
                                  self.sts_makeword(data[2], data[3])) if (result == COMM_SUCCESS) else 0
        return data_read, result, error

    def writeTxOnly(self, sts_id, address, length, data):
        txpacket = [0] * (length + 7)

        txpacket[PKT_ID] = sts_id
        txpacket[PKT_LENGTH] = length + 3
        txpacket[PKT_INSTRUCTION] = INST_WRITE
        txpacket[PKT_PARAMETER0] = address

        txpacket[PKT_PARAMETER0 + 1: PKT_PARAMETER0 + 1 + length] = data[0: length]

        result = self.txPacket(txpacket)
        self.portHandler.is_using = False

        return result

    def writeTxRx(self, sts_id, address, length, data):
        txpacket = [0] * (length + 7)

        txpacket[PKT_ID] = sts_id
        txpacket[PKT_LENGTH] = length + 3
        txpacket[PKT_INSTRUCTION] = INST_WRITE
        txpacket[PKT_PARAMETER0] = address

        txpacket[PKT_PARAMETER0 + 1: PKT_PARAMETER0 + 1 + length] = data[0: length]
        rxpacket, result, error = self.txRxPacket(txpacket)

        return result, error

    def write1ByteTxOnly(self, sts_id, address, data):
        data_write = [data]
        return self.writeTxOnly(sts_id, address, 1, data_write)

    def write1ByteTxRx(self, sts_id, address, data):
        data_write = [data]
        return self.writeTxRx(sts_id, address, 1, data_write)

    def write2ByteTxOnly(self, sts_id, address, data):
        data_write = [self.sts_lobyte(data), self.sts_hibyte(data)]
        return self.writeTxOnly(sts_id, address, 2, data_write)

    def write2ByteTxRx(self, sts_id, address, data):
        data_write = [self.sts_lobyte(data), self.sts_hibyte(data)]
        return self.writeTxRx(sts_id, address, 2, data_write)

    def write4ByteTxOnly(self, sts_id, address, data):
        data_write = [self.sts_lobyte(self.sts_loword(data)),
                      self.sts_hibyte(self.sts_loword(data)),
                      self.sts_lobyte(self.sts_hiword(data)),
                      self.sts_hibyte(self.sts_hiword(data))]
        return self.writeTxOnly(sts_id, address, 4, data_write)

    def write4ByteTxRx(self, sts_id, address, data):
        data_write = [self.sts_lobyte(self.sts_loword(data)),
                      self.sts_hibyte(self.sts_loword(data)),
                      self.sts_lobyte(self.sts_hiword(data)),
                      self.sts_hibyte(self.sts_hiword(data))]
        return self.writeTxRx(sts_id, address, 4, data_write)

    def regWriteTxOnly(self, sts_id, address, length, data):
        txpacket = [0] * (length + 7)

        txpacket[PKT_ID] = sts_id
        txpacket[PKT_LENGTH] = length + 3
        txpacket[PKT_INSTRUCTION] = INST_REG_WRITE
        txpacket[PKT_PARAMETER0] = address

        txpacket[PKT_PARAMETER0 + 1: PKT_PARAMETER0 + 1 + length] = data[0: length]

        result = self.txPacket(txpacket)
        self.portHandler.is_using = False

        return result

    def regWriteTxRx(self, sts_id, address, length, data):
        txpacket = [0] * (length + 7)

        txpacket[PKT_ID] = sts_id
        txpacket[PKT_LENGTH] = length + 3
        txpacket[PKT_INSTRUCTION] = INST_REG_WRITE
        txpacket[PKT_PARAMETER0] = address

        txpacket[PKT_PARAMETER0 + 1: PKT_PARAMETER0 + 1 + length] = data[0: length]

        _, result, error = self.txRxPacket(txpacket)

        return result, error

    def syncReadTx(self, start_address, data_length, param, param_length):
        txpacket = [0] * (param_length + 8)
        # 8: HEADER0 HEADER1 ID LEN INST START_ADDR DATA_LEN CHKSUM

        txpacket[PKT_ID] = BROADCAST_ID
        txpacket[PKT_LENGTH] = param_length + 4  # 7: INST START_ADDR DATA_LEN CHKSUM
        txpacket[PKT_INSTRUCTION] = INST_SYNC_READ
        txpacket[PKT_PARAMETER0 + 0] = start_address
        txpacket[PKT_PARAMETER0 + 1] = data_length

        txpacket[PKT_PARAMETER0 + 2: PKT_PARAMETER0 + 2 + param_length] = param[0: param_length]

        # print(txpacket)
        result = self.txPacket(txpacket)
        return result

    def syncReadRx(self, data_length, param_length):
        wait_length = (6 + data_length) * param_length
        self.portHandler.setPacketTimeout(wait_length)
        rxpacket = []
        rx_length = 0
        while True:
            rxpacket.extend(self.portHandler.readPort(wait_length - rx_length))
            rx_length = len(rxpacket)
            if rx_length >= wait_length:
                result = COMM_SUCCESS
                break
            else:
                # check timeout
                if self.portHandler.isPacketTimeout():
                    if rx_length == 0:
                        result = COMM_RX_TIMEOUT
                    else:
                        result = COMM_RX_CORRUPT
                    break
        self.portHandler.is_using = False
        return result, rxpacket

    def syncWriteTxOnly(self, start_address, data_length, param, param_length):
        txpacket = [0] * (param_length + 8)
        # 8: HEADER0 HEADER1 ID LEN INST START_ADDR DATA_LEN ... CHKSUM

        txpacket[PKT_ID] = BROADCAST_ID
        txpacket[PKT_LENGTH] = param_length + 4  # 4: INST START_ADDR DATA_LEN ... CHKSUM
        txpacket[PKT_INSTRUCTION] = INST_SYNC_WRITE
        txpacket[PKT_PARAMETER0 + 0] = start_address
        txpacket[PKT_PARAMETER0 + 1] = data_length

        txpacket[PKT_PARAMETER0 + 2: PKT_PARAMETER0 + 2 + param_length] = param[0: param_length]

        _, result, _ = self.txRxPacket(txpacket)

        return result



import time
import serial
import sys

from .values import *


class PortHandler(object):
    def __init__(self, port_name):
        self.is_open = False
        self.baudrate = DEFAULT_BAUDRATE
        self.packet_start_time = 0.0
        self.packet_timeout = 0.0
        self.tx_time_per_byte = 0.0

        self.is_using = False
        self.port_name = port_name
        self.ser = None

    def openPort(self):
        return self.setupPort()

    def closePort(self):
        self.ser.close()
        self.is_open = False

    def clearPort(self):
        self.ser.flush()

    def setPortName(self, port_name):
        self.port_name = port_name

    def getPortName(self):
        return self.port_name

    def getBaudRate(self):
        return self.baudrate

    def getBytesAvailable(self):
        return self.ser.in_waiting

    def readPort(self, length):
        if (sys.version_info > (3, 0)):
            return self.ser.read(length)
        else:
            return [ord(ch) for ch in self.ser.read(length)]

    def writePort(self, packet):
        return self.ser.write(packet)

    def setPacketTimeout(self, packet_length):
        self.packet_start_time = self.getCurrentTime()
        self.packet_timeout = (self.tx_time_per_byte * packet_length) + (self.tx_time_per_byte * 3.0) + LATENCY_TIMER

    def setPacketTimeoutMillis(self, msec):
        self.packet_start_time = self.getCurrentTime()
        self.packet_timeout = msec

    def isPacketTimeout(self):
        if self.getTimeSinceStart() > self.packet_timeout:
            self.packet_timeout = 0
            return True

        return False

    def getCurrentTime(self):
        return round(time.time() * 1000000000) / 1000000.0

    def getTimeSinceStart(self):
        time_since = self.getCurrentTime() - self.packet_start_time
        if time_since < 0.0:
            self.packet_start_time = self.getCurrentTime()

        return time_since

    def setupPort(self):
        if self.is_open:
            self.closePort()

        self.ser = serial.Serial(
            port=self.port_name,
            baudrate=self.baudrate,
            bytesize=serial.EIGHTBITS,
            timeout=0
        )

        self.is_open = True

        self.ser.reset_input_buffer()

        self.tx_time_per_byte = (1000.0 / self.baudrate) * 10.0

        return True




DEFAULT_BAUDRATE = 1000000
LATENCY_TIMER = 50 

TXPACKET_MAX_LEN = 250
RXPACKET_MAX_LEN = 250

MIN_POSITION = 0
MAX_POSITION = 4095

MAX_SPEED = 3400

MAX_CORRECTION = 2047

# for Protocol Packet
PKT_HEADER_0 = 0
PKT_HEADER_1 = 1
PKT_ID = 2
PKT_LENGTH = 3
PKT_INSTRUCTION = 4
PKT_ERROR = 4
PKT_PARAMETER0 = 5

# Protocol Error bit
ERRBIT_VOLTAGE = 1
ERRBIT_ANGLE = 2
ERRBIT_OVERHEAT = 4
ERRBIT_OVERELE = 8
ERRBIT_OVERLOAD = 32


BROADCAST_ID = 0xFE  # 254
MAX_ID = 0xFC  # 252
STS_END = 0

# Instruction for STS Protocol
INST_PING = 1
INST_READ = 2
INST_WRITE = 3
INST_REG_WRITE = 4
INST_ACTION = 5
INST_SYNC_WRITE = 131  # 0x83
INST_SYNC_READ = 130  # 0x82

# Communication Result
COMM_SUCCESS = 0  # tx or rx packet communication success
COMM_PORT_BUSY = -1  # Port is busy (in use)
COMM_TX_FAIL = -2  # Failed transmit instruction packet
COMM_RX_FAIL = -3  # Failed get status packet
COMM_TX_ERROR = -4  # Incorrect instruction packet
COMM_RX_WAITING = -5  # Now recieving status packet
COMM_RX_TIMEOUT = -6  # There is no status packet
COMM_RX_CORRUPT = -7  # Incorrect status packet
COMM_NOT_AVAILABLE = -9  #



# Bus Speed
STS_1M = 0
STS_0_5M = 1
STS_250K = 2
STS_128K = 3
STS_115200 = 4
STS_76800 = 5
STS_57600 = 6
STS_38400 = 7

# EPROM RO
STS_MODEL_L = 3
STS_MODEL_H = 4

# EPROM RW
STS_ID = 5
STS_BAUD_RATE = 6
STS_MIN_ANGLE_LIMIT_L = 9
STS_MIN_ANGLE_LIMIT_H = 10
STS_MAX_ANGLE_LIMIT_L = 11
STS_MAX_ANGLE_LIMIT_H = 12
STS_CW_DEAD = 26
STS_CCW_DEAD = 27
STS_OFS_L = 31
STS_OFS_H = 32
STS_MODE = 33

# SRAM RW
STS_TORQUE_ENABLE = 40
STS_ACC = 41
STS_GOAL_POSITION_L = 42
STS_GOAL_POSITION_H = 43
STS_GOAL_TIME_L = 44
STS_GOAL_TIME_H = 45
STS_GOAL_SPEED_L = 46
STS_GOAL_SPEED_H = 47
STS_LOCK = 55

# SRAM RO
STS_PRESENT_POSITION_L = 56
STS_PRESENT_POSITION_H = 57
STS_PRESENT_SPEED_L = 58
STS_PRESENT_SPEED_H = 59
STS_PRESENT_LOAD_L = 60
STS_PRESENT_LOAD_H = 61
STS_PRESENT_VOLTAGE = 62
STS_PRESENT_TEMPERATURE = 63
STS_MOVING = 66
STS_PRESENT_CURRENT_L = 69
STS_PRESENT_CURRENT_H = 70





from .values import *

class GroupSyncWrite:
    def __init__(self, ph, start_address, data_length):
        self.ph = ph
        self.start_address = start_address
        self.data_length = data_length

        self.is_param_changed = False
        self.param = []
        self.data_dict = {}

        self.clearParam()

    def makeParam(self):
        if not self.data_dict:
            return

        self.param = []

        for sts_id in self.data_dict:
            if not self.data_dict[sts_id]:
                return

            self.param.append(sts_id)
            self.param.extend(self.data_dict[sts_id])

    def addParam(self, sts_id, data):
        if sts_id in self.data_dict:  # sts_id already exist
            return False

        if len(data) > self.data_length:  # input data is longer than set
            return False

        self.data_dict[sts_id] = data

        self.is_param_changed = True
        return True

    def removeParam(self, sts_id):
        if sts_id not in self.data_dict:  # NOT exist
            return

        del self.data_dict[sts_id]

        self.is_param_changed = True

    def changeParam(self, sts_id, data):
        if sts_id not in self.data_dict:  # NOT exist
            return False

        if len(data) > self.data_length:  # input data is longer than set
            return False

        self.data_dict[sts_id] = data

        self.is_param_changed = True
        return True

    def clearParam(self):
        self.data_dict.clear()

    def txPacket(self):
        if len(self.data_dict.keys()) == 0:
            return COMM_NOT_AVAILABLE

        if self.is_param_changed is True or not self.param:
            self.makeParam()

        return self.ph.syncWriteTxOnly(self.start_address, self.data_length, self.param,
                                       len(self.data_dict.keys()) * (1 + self.data_length))



import threading
import time

from .port_handler import *
from .protocol_packet_handler import *
from .group_sync_write import *
from .group_sync_read import *
from .values import *


__all__ = ['ST3215']


class ST3215(protocol_packet_handler):

    def __init__(self, device):

        self.portHandler = PortHandler(device)
        
        if not self.portHandler.openPort():
            raise ValueError(f"Could not open port: {device}")

        protocol_packet_handler.__init__(self, self.portHandler)

        self.groupSyncWrite = GroupSyncWrite(self, STS_ACC, 7)
        self.lock = threading.Lock()


    def PingServo(self, sts_id):
        """
        Check the presence of a servo.

        :param sts_id: Servo ID

        :return: True in case of success otherwise False
        """
        model, comm, error = self.ping(sts_id)
        if comm != COMM_SUCCESS or model == 0 or error != 0:
            return False
        return True


    def ListServos(self):
        """
        Scan the bus to determine all servo present

        :return: A list of servo ID
        """
        servos=[]
        for id in range(0, 254):
            if self.PingServo(id):
                servos.append(id)

        return servos


    def ReadLoad(self, sts_id):
        """
        Load of the servo. 
        Load value is determined by: The voltage duty cycle of the current control output driving motor

        :param sts_id: Servo ID

        :return: Load value in percentage. None in case of error.
        """
        load, comm, error = self.read1ByteTxRx(sts_id, STS_PRESENT_LOAD_L)
        if comm == 0 and error == 0:
            return load * 0.1
        else:
            return None

    def ReadVoltage(self, sts_id):
        """
        Current Voltage of the servo. 

        :param sts_id: Servo ID

        :return: Current Voltage in V. None in case of error.
        """
        voltage, comm, error = self.read1ByteTxRx(sts_id, STS_PRESENT_VOLTAGE)
        if comm == 0 and error == 0:
            return voltage * 0.1
        else:
            return None

    def ReadCurrent(self, sts_id):
        """
        Current current of the servo. 

        :param sts_id: Servo ID

        :return: Current current in mA. None in case of error.
        """
        current, comm, error = self.read1ByteTxRx(sts_id, STS_PRESENT_CURRENT_L)
        if comm == 0 and error == 0:
            return current * 6.5
        else:
            return None

    def ReadTemperature(self, sts_id):
        """
        Current temperature of the servo. 

        :param sts_id: Servo ID

        :return: Current temperature in °C. None in case of error.
        """
        temperature, comm, error =  self.read1ByteTxRx(sts_id, STS_PRESENT_TEMPERATURE)
        if comm == 0 and error == 0:
            return temperature
        else:
            return None


    def ReadAccelaration(self, sts_id):
        """
        Current value of the acceleration of the servo. 

        :param sts_id: Servo ID

        :return: Current acceleration value. None in case of error.
        """
        acc, comm, error =  self.read1ByteTxRx(sts_id, STS_ACC)
        if comm == 0 and error == 0:
            return acc
        else:
            return None


    def ReadMode(self, sts_id):
        """
        Current mode of the servo. 
          - 0: Position Mode
          - 1: Constant speed mode
          - 2: PWM Mode
          - 3: Step servo mode

        :param sts_id: Servo ID

        :return: Current mode. None in case of error.
        """
        mode, comm, error =  self.read1ByteTxRx(sts_id, STS_MODE)
        if comm == 0 and error == 0:
            return mode
        else:
            return None


    def ReadCorrection(self, sts_id):
        """
        Current value of position correction for the servo. 

        :param sts_id: Servo ID

        :return: Current correction value. None in case of error.
        """
        correction, comm, error =  self.read2ByteTxRx(sts_id, STS_OFS_L)
        if comm == 0 and error == 0:
            mask = 0x07FFF
            bits = correction & mask
            if (correction & 0x0800) != 0:
                bits = -1 * (bits & 0x7FF)
            return bits
        else:
            return None


    def IsMoving(self, sts_id):
        """
        Is the servo moving

        :param sts_id: Servo ID

        :return: True is the servo is moving otherwise False. None in case of error.
        """
        moving, comm, error =  self.read1ByteTxRx(sts_id, STS_MOVING)
        if comm == 0 and error == 0:
            return bool(moving)
        else:
            return None


    def SetAcceleration(self, sts_id, acc):
        """
        Configure the Acceleration value for the servo

        :param sts_id: Servo ID
        :param acc: Acceleration value (0-254). Unit: 100 step/s^2)

        :return: True if the configuration a been succesfully set. None in case of error.
        """
        txpacket = [acc]
        comm, error = self.writeTxRx(sts_id, STS_ACC, len(txpacket), txpacket)
        if comm == 0 and error == 0:
            return True
        else:
            return None


    def SetSpeed(self, sts_id, speed):
        """
        Configure the Speed value for the servo

        :param sts_id: Servo ID
        :param acc: Speed value (0-3400). Unit: Step/s

        :return: True if the configuration a been succesfully set. None in case of error.
        """
        txpacket = [self.sts_lobyte(speed), self.sts_hibyte(speed)]
        comm, error = self.writeTxRx(sts_id, STS_GOAL_SPEED_L, len(txpacket), txpacket)
        if comm == 0 and error == 0:
            return True
        else:
            return None


    def StopServo(self, sts_id):
        """
        Stop the servo (Set torque to 0)

        :param sts_id: Servo ID

        :return: True if the configuration a been succesfully set. None in case of error.
        """
        txpacket = [0]
        comm, error = self.writeTxRx(sts_id, STS_TORQUE_ENABLE, len(txpacket), txpacket)
        if comm == 0 and error == 0:
            return True
        else:
            return None


    def StartServo(self, sts_id):
        """
        Start the servo (Set torque to 1)

        :param sts_id: Servo ID

        :return: True if the configuration a been succesfully set. None in case of error.
        """
        txpacket = [1]
        return self.writeTxRx(sts_id, STS_TORQUE_ENABLE, len(txpacket), txpacket)



    def SetMode(self, sts_id, mode):
        """
        Configure the operational mode for the servo (Position, rotating, PWM, step)

        :param sts_id: Servo ID
        :param mode: Mode ID (0, 1, 2 or 3 - Cf register values)

        :return: True if the configuration a been succesfully set. None in case of error.
        """
        txpacket = [mode]
        return self.writeTxRx(sts_id, STS_MODE, len(txpacket), txpacket)



    def CorrectPosition(self, sts_id, correction):
        """
        Add a position correction

        :param sts_id: Servo ID
        :param correction: correction (in steps, could be negative)

        :return: True if the configuration a been succesfully set. None in case of error.
        """
        corr = abs(correction)
        if corr > MAX_CORRECTION:
            corr = MAX_CORRECTION

        txpacket = [self.sts_lobyte(corr), self.sts_hibyte(corr)]

        if correction < 0:
            txpacket[1] |= (1 << 3)

        return self.writeTxRx(sts_id, STS_OFS_L, len(txpacket), txpacket)


    def Rotate(self, sts_id, speed):
        """
        Start rotating

        :param sts_id: Servo ID
        :param speed: servo speed (could be negative, if so rotate counterwise)

        :return: True if the configuration a been succesfully set. None in case of error.
        """

        self.SetMode(sts_id, 1)

        abs_speed = abs(speed)
        if abs_speed > MAX_SPEED:
            abs_speed = MAX_SPEED

        txpacket = [self.sts_lobyte(abs_speed), self.sts_hibyte(abs_speed)]

        if speed < 0:
            txpacket[1] |= (1 << 7)

        return self.writeTxRx(sts_id, STS_GOAL_SPEED_L, len(txpacket), txpacket)



    def getBlockPosition(self, sts_id):
        """
        Get the next blocking position

        :param sts_id: Servo ID

        :return: blocking position. None in case of error.
        """

        stop_matches = 0
        while True:
            moving = self.IsMoving(sts_id)
            if moving == None:
                self.SetMode(sts_id, 0)
                self.StopServo(sts_id)
                return None

            if moving == False:
                position = self.ReadPosition(sts_id)
                self.SetMode(sts_id, 0)
                self.StopServo(sts_id)

                if position == None:
                    return None
                else:
                    stop_matches += 1
                    if stop_matches > 4:
                        return position
            else:
                stop_matches = 0

            time.sleep(0.02)



    def DefineMiddle(self, sts_id):
        """
        Define the 2048 position (Set torque to 128)

        :param sts_id: Servo ID

        :return: True if the configuration a been succesfully set. None in case of error.
        """
        txpacket = [128]
        comm, error = self.writeTxRx(sts_id, STS_TORQUE_ENABLE, len(txpacket), txpacket)
        if comm == 0 and error == 0:
            return True
        else:
            return None



    def TareServo(self, sts_id):
        """
        Tare a Servo: Find its min and max position, then configure the new middle value (MAX_VALUE / 2 - 2048 by default) to the middle of the servo course

        WARNING: Must only be use for a servo that has at least one blocking position. Never use this function for a free rotation servo.

        :param sts_id: Servo ID

        :return: min and max position in the servo course. None in case of error.
        """

        self.SetAcceleration(sts_id, 100)
        self.Rotate(sts_id, 250)
        time.sleep(0.5)

        max_position = self.getBlockPosition(sts_id)


        self.Rotate(sts_id, -250)
        time.sleep(0.5)

        min_position = self.getBlockPosition(sts_id)

        if min_position is not None and max_position is not None:

            # Now, set the middle of the path to 2048
            if min_position >= max_position:
                distance = int(((MAX_POSITION - min_position + max_position) / 2))
                middle = (min_position + distance) % MAX_POSITION
            else:
                distance = int(((max_position - min_position) / 2))
                middle = min_position + distance


            if self.MoveTo(sts_id, middle) is not None:
                time.sleep(2)
                if self.DefineMiddle(sts_id) is not None:
                    if self.CorrectPosition(sts_id, -2048) is not None:
                        min_position = 0
                        max_position = distance * 2

                self.MoveTo(sts_id, distance)


        return min_position, max_position



    def MoveTo(self, sts_id, position, speed = 2400, acc = 50):
        """
        Move the servo to a pre defined position

        :param sts_id: Servo ID

        :return: True. None in case of error.
        """

        res_mode = self.SetMode(sts_id, 0) == None
        res_acc = self.SetAcceleration(sts_id, acc)
        res_speed = self.SetSpeed(sts_id, speed)

        if res_acc == None or res_speed == None or res_mode:
            return None

        return self.WritePosition(sts_id, position)



    def WritePosition(self, sts_id, position):
        txpacket = [self.sts_lobyte(position), self.sts_hibyte(position)]
        comm, error = self.writeTxRx(sts_id, STS_GOAL_POSITION_L, len(txpacket), txpacket)
        if comm == 0 and error == 0:
            return True
        else:
            return None


    def ReadPosition(self, sts_id):
        """
        Get the current position

        :param sts_id: Servo ID

        :return: position in case of success, otherwise None
        """
        position, comm, error = self.read2ByteTxRx(sts_id, STS_PRESENT_POSITION_L)
        if comm == 0 and error == 0:
            return position
        else:
            return None

    def ReadSpeed(self, sts_id):
        """
        Get the current speed

        :param sts_id: Servo ID

        :return: speed in case of success, otherwise None
        """
        sts_present_speed, sts_comm_result, sts_error = self.read2ByteTxRx(sts_id, STS_PRESENT_SPEED_L)
        return self.sts_tohost(sts_present_speed, 15), sts_comm_result, sts_error



    def LockEprom(self, sts_id):
        """
        Lock the servo Eeprom.

        :param sts_id: Servo ID

        :return: 0 in case of success
        """
        return self.write1ByteTxOnly(sts_id, STS_LOCK, 1)

    def UnLockEprom(self, sts_id):
        """
        Unlock the servo Eeprom.

        :param sts_id: Servo ID

        :return: 0 in case of success
        """
        return self.write1ByteTxOnly(sts_id, STS_LOCK, 0)

    def ChangeId(self, sts_id, new_id):
        """
        Change ID of a Servo.

        :param sts_id: Actual ID for the servo (1 for a brand new servo)
        :param new_id: New ID for the servo

        :return: None when sucedeed otherwise the error message 
        """
        if isinstance(new_id, int) and 0 <= new_id <= 253:
            if not self.PingServo(sts_id):
                return f"Could not find servo: {sts_id}" 

            if self.UnLockEprom(sts_id) != COMM_SUCCESS:
                return "Could not unlock Eprom" 

            if self.write1ByteTxOnly(sts_id, STS_ID, new_id) != COMM_SUCCESS:
                return "Could not change Servo ID" 

            self.LockEprom(sts_id)
            return None
        else:
            return "new_id is not between 0 and 253" 






#!/usr/bin/env python
"""st3215 extension builder and installer"""

import io
import sys

import setuptools

name = "st3215"
desc = "A library to manage ST3215 servo motors"

with io.open("README.md", encoding="utf-8") as strm:
    long_desc = strm.read()

classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: Freely Distributable",
    "Operating System :: OS Independent",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: Implementation",
    "Programming Language :: Python :: Implementation :: CPython",
]
author = "Mickael Roger"
author_email = "mickael@mickael-roger.com"
url = "https://github.com/Mickael-Roger/python-st3215"
license = "Apache License Version 2.0"
packages = ["st3215"]

needs_pytest = set(["pytest", "test"]).intersection(sys.argv)
pytest_runner = ["pytest_runner"] if needs_pytest else []

test_require = []
with io.open("requirements.txt") as f:
    test_require = [l.strip() for l in f if not l.startswith("#")]

install_require = []
with io.open("requirements.txt") as f:
    install_require = [l.strip() for l in f if not l.startswith("#")]

setup_params = dict(
    name=name,
    version="0.1.3",
    description=desc,
    long_description=long_desc,
    classifiers=classifiers,
    author=author,
    author_email=author_email,
    url=url,
    license=license,
    packages=packages,
    include_package_data=True,
    install_requires=install_require,
    tests_require=test_require,
    setup_requires=pytest_runner,
    python_requires=">=3.10",
)


def main():
    """Package installation entry point."""
    setuptools.setup(**setup_params)


if __name__ == "__main__":
    main()



from .values import *

class GroupSyncRead:
    def __init__(self, ph, start_address, data_length):
        self.ph = ph
        self.start_address = start_address
        self.data_length = data_length

        self.last_result = False
        self.is_param_changed = False
        self.param = []
        self.data_dict = {}

        self.clearParam()

    def makeParam(self):
        if not self.data_dict:  # len(self.data_dict.keys()) == 0:
            return

        self.param = []

        for scs_id in self.data_dict:
            self.param.append(scs_id)

    def addParam(self, sts_id):
        if sts_id in self.data_dict:  # sts_id already exist
            return False

        self.data_dict[sts_id] = []  # [0] * self.data_length

        self.is_param_changed = True
        return True

    def removeParam(self, sts_id):
        if sts_id not in self.data_dict:  # NOT exist
            return

        del self.data_dict[sts_id]

        self.is_param_changed = True

    def clearParam(self):
        self.data_dict.clear()

    def txPacket(self):
        if len(self.data_dict.keys()) == 0:

            return COMM_NOT_AVAILABLE

        if self.is_param_changed is True or not self.param:
            self.makeParam()

        return self.ph.syncReadTx(self.start_address, self.data_length, self.param, len(self.data_dict.keys()))

    def rxPacket(self):
        self.last_result = True

        result = COMM_RX_FAIL

        if len(self.data_dict.keys()) == 0:
            return COMM_NOT_AVAILABLE

        result, rxpacket = self.ph.syncReadRx(self.data_length, len(self.data_dict.keys()))
        # print(rxpacket)
        if len(rxpacket) >= (self.data_length+6):
            for sts_id in self.data_dict:
                self.data_dict[sts_id], result = self.readRx(rxpacket, sts_id, self.data_length)
                if result != COMM_SUCCESS:
                    self.last_result = False
                # print(sts_id)
        else:
            self.last_result = False
        # print(self.last_result)
        return result

    def txRxPacket(self):
        result = self.txPacket()
        if result != COMM_SUCCESS:
            return result

        return self.rxPacket()

    def readRx(self, rxpacket, sts_id, data_length):
        # print(sts_id)
        # print(rxpacket)
        data = []
        rx_length = len(rxpacket)
        # print(rx_length)
        rx_index = 0;
        while (rx_index+6+data_length) <= rx_length:
            headpacket = [0x00, 0x00, 0x00]
            while rx_index < rx_length:
                headpacket[2] = headpacket[1];
                headpacket[1] = headpacket[0];
                headpacket[0] = rxpacket[rx_index];
                rx_index += 1
                if (headpacket[2] == 0xFF) and (headpacket[1] == 0xFF) and headpacket[0] == sts_id:
                    # print(rx_index)
                    break
            # print(rx_index+3+data_length)
            if (rx_index+3+data_length) > rx_length:
                break;
            if rxpacket[rx_index] != (data_length+2):
                rx_index += 1
                # print(rx_index)
                continue
            rx_index += 1
            Error = rxpacket[rx_index]
            rx_index += 1
            calSum = sts_id + (data_length+2) + Error
            data = [Error]
            data.extend(rxpacket[rx_index : rx_index+data_length])
            for i in range(0, data_length):
                calSum += rxpacket[rx_index]
                rx_index += 1
            calSum = ~calSum & 0xFF
            # print(calSum)
            if calSum != rxpacket[rx_index]:
                return None, COMM_RX_CORRUPT
            return data, COMM_SUCCESS 
        # print(rx_index)
        return None, COMM_RX_CORRUPT

    def isAvailable(self, sts_id, address, data_length):
        #if self.last_result is False or sts_id not in self.data_dict:
        if sts_id not in self.data_dict:
            return False, 0

        if (address < self.start_address) or (self.start_address + self.data_length - data_length < address):
            return False, 0
        if not self.data_dict[sts_id]:
            return False, 0
        if len(self.data_dict[sts_id])<(data_length+1):
            return False, 0
        return True, self.data_dict[sts_id][0]

    def getData(self, sts_id, address, data_length):
        if data_length == 1:
            return self.data_dict[sts_id][address-self.start_address+1]
        elif data_length == 2:
            return self.ph.scs_makeword(self.data_dict[sts_id][address-self.start_address+1],
                                self.data_dict[sts_id][address-self.start_address+2])
        elif data_length == 4:
            return self.ph.scs_makedword(self.ph.scs_makeword(self.data_dict[sts_id][address-self.start_address+1],
                                              self.data_dict[sts_id][address-self.start_address+2]),
                                 self.ph.scs_makeword(self.data_dict[sts_id][address-self.start_address+3],
                                              self.data_dict[sts_id][address-self.start_address+4]))
        else:
            return 0
